

///////////////////////////////////////////////////////////////////////////////
//
//	level.cpp
//

#include "id.h"
#include "level_select.h"
#include "level_setting.h"
#include "project.h"

#pragma warning(disable: 4355)


LPCSTR LevelManager::pAppNameMapView = "MapView";
LPCSTR LevelManager::pClassName = "sdk2ed_mapview";


BOOL LevelManager::Init(){
	if (!register_class_ex(
		0,
		(HICON)NULL,
		(HICON)NULL,
		LoadCursor(NULL, IDC_ARROW),
		(HBRUSH)GetStockObject(NULL_BRUSH),
		pClassName,
		NULL
		)
	){
		return FALSE;
	}
	if (!LevelEditInspector::Init()) return FALSE;
	return TRUE;
}

LevelManager::LevelManager(ProjectManager *_project)
: project(_project), buffer(),
mpan(this, 10, 0x0001), inspector(_project), recent_levels(this, _project),
tilemap(this, _project), sprites(this, _project), bananas(this, _project), cameras(this, _project),
mouse_drag(DRAG_NONE)
{
}

LevelManager::~LevelManager(){
}


BOOL LevelManager::create(){
	level_num		= GetPrivateProfileInt(pAppNameMapView, "LevelNumber", 0x25, ConfigFileName);
	if (!create_window_ex(pClassName, "",
		WS_CHILD | WS_VISIBLE | WS_HSCROLL | WS_VSCROLL,
		WS_EX_CLIENTEDGE,
		0, 0, 0, 0,
		project))
	{
		return FALSE;
	}
	TabPage *(pages[3]);
	pages[0] = &sprites;
	pages[1] = &bananas;
	pages[2] = &cameras;
	if (!inspector.create(pages, 3)) return FALSE;
	return TRUE;
}

void LevelManager::save_config(){
	char s[16];
	sprintf_s(s, sizeof s, "%d", level_num);
	WritePrivateProfileString(pAppNameMapView, "LevelNumber", s, ConfigFileName);
	recent_levels.save_data();
	tilemap.save_data();
	sprites.save_data();
	bananas.save_data();
	cameras.save_data();
	return;
}

LRESULT LevelManager::Proc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam){
	PAINTSTRUCT ps;
	HDC hdc;
	POINT pt;
	switch (msg){
		case WM_CREATE:
			hdc = GetDC(hWnd);
			if (!buffer.create(hdc, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN))){
				ReleaseDC(hWnd, hdc);
				return -1;
			}
			if (!recent_levels.create()) return -1;
			if (!tilemap.create()) return -1;
			if (!sprites.create_sys()) return -1;
			if (!bananas.create_sys()) return -1;
			if (!cameras.create_sys()) return -1;
			ReleaseDC(hWnd, hdc);
			set_range_h(0, 0);
			set_page_size_h(1);
			set_line_size_h(32);
			set_position_h(0);
			set_range_v(0, 0);
			set_page_size_v(1);
			set_line_size_v(32);
			set_position_v(0);
			return 0;
		case WM_LBUTTONDOWN:
			clicked((SHORT)LOWORD(lParam), (SHORT)HIWORD(lParam));
			return 0;
		case WM_LBUTTONUP:
			unclicked((SHORT)LOWORD(lParam), (SHORT)HIWORD(lParam));
			return 0;
		case WM_MOUSEMOVE:
			if (mouse_drag != DRAG_NONE){
				dragging((SHORT)LOWORD(lParam), (SHORT)HIWORD(lParam));
			}
			pt.x = (SHORT)LOWORD(lParam);
			pt.y = (SHORT)HIWORD(lParam);
			mpan.handle_mousemove(pt);
			return 0;
		case WM_SETCURSOR:
			if (mouse_drag == DRAG_NONE){
				WORD hit = LOWORD(lParam);
				if (hit == HTCLIENT){
					POINT pt;
					GetCursorPos(&pt);
					ScreenToClient(hWnd, &pt);
					mouseover(pt.x, pt.y);
				}else if (hit == HTHSCROLL || hit == HTVSCROLL){
					SetCursor(LoadCursor(NULL, IDC_ARROW));
				}
			}
			return 0;
		case WM_MBUTTONDOWN:
			pt.x = (SHORT)LOWORD(lParam);
			pt.y = (SHORT)HIWORD(lParam);
			mpan.start(pt);
			return 0;
		case WM_TIMER:
			if (mpan.handle_timer(wParam)) return 0;
			break;
		case WM_MBUTTONUP:
			mpan.stop();
			return 0;
		case WM_HSCROLL:
			handle_message_h(wParam, lParam);
			return 0;
		case WM_VSCROLL:
			handle_message_v(wParam, lParam);
			return 0;
		case WM_CHAR:
			switch (wParam){
				case 'm': command(IDM_VIEW_TILEMAP, 0L); return 0;
				case 's': command(IDM_VIEW_SPRITES, 0L); return 0;
				case 'b': command(IDM_VIEW_BANANAS, 0L); return 0;
				case 'c': command(IDM_VIEW_CAMERAS, 0L); return 0;
			}
			break;
		case WM_KEYDOWN:
			switch (wParam){
				case VK_UP:
					if (GetKeyState(VK_CONTROL) < 0){
						scroll_v(-256);
					}else{
						SendMessage(hWnd, WM_VSCROLL, SB_LINEUP, 0);
					}
					return 0;
				case VK_DOWN:
					if (GetKeyState(VK_CONTROL) < 0){
						scroll_v(256);
					}else{
						SendMessage(hWnd, WM_VSCROLL, SB_LINEDOWN, 0);
					}
					return 0;
				case VK_LEFT:
					if (GetKeyState(VK_CONTROL) < 0){
						scroll_h(-256);
					}else{
						SendMessage(hWnd, WM_HSCROLL, SB_LINELEFT, 0);
					}
					return 0;
				case VK_RIGHT:
					if (GetKeyState(VK_CONTROL) < 0){
						scroll_h(256);
					}else{
						SendMessage(hWnd, WM_HSCROLL, SB_LINERIGHT, 0);
					}
					return 0;
			}
			break;
		case WM_MOUSEWHEEL:
			if (map_vert){
				SendMessage(hWnd, WM_VSCROLL, ((SHORT)HIWORD(wParam))>0 ? SB_LINEUP : SB_LINEDOWN, 0);
			}else{
				SendMessage(hWnd, WM_HSCROLL, ((SHORT)HIWORD(wParam))>0 ? SB_LINELEFT : SB_LINERIGHT, 0);
			}
			return 0;
		case WM_SIZE:
			resize_h(LOWORD(lParam));
			resize_v(HIWORD(lParam));
			return 0;
		case WM_PAINT:
			hdc = BeginPaint(hWnd, &ps);
			{
				RECT rt = ps.rcPaint;
				int pos_h = get_pos_h(), pos_v = get_pos_v();
				rt.left += pos_h, rt.right += pos_h;
				rt.top += pos_v, rt.bottom += pos_v;
				redraw(&rt);
				BitBlt(hdc, ps.rcPaint.left, ps.rcPaint.top, ps.rcPaint.right-ps.rcPaint.left,
					ps.rcPaint.bottom-ps.rcPaint.top, buffer.get_HDC(), 0, 0, SRCCOPY);
			}
			EndPaint(hWnd, &ps);
			return 0;
		case WM_DESTROY:
			save_config();
			buffer.destroy();
			inspector.destroy();
			return 0;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}

void LevelManager::scrolled_h(int nNewPos, UINT nNewPage, int nOldPos, UINT nOldPage){
	int ndx = nOldPos - nNewPos;
	UINT height = get_page_v();
	RECT rtScroll = {0, 0, nOldPage, height};
	ScrollWindowEx(hWnd, ndx, 0, &rtScroll, NULL, NULL, NULL, SW_INVALIDATE);
	if (ndx > 0){
		RECT rtLeft = {0, 0, ndx, height};
		InvalidateRect(hWnd, &rtLeft, TRUE);
	}
	if (nOldPage+ndx < (signed)nNewPage){
		RECT rtRight = {nOldPage+ndx, 0, nNewPage, height};
		InvalidateRect(hWnd, &rtRight, TRUE);
	}
	return;
}

void LevelManager::scrolled_v(int nNewPos, UINT nNewPage, int nOldPos, UINT nOldPage){
	int ndy = nOldPos - nNewPos;
	UINT width = get_page_h();
	RECT rtScroll = {0, 0, width, nOldPage};
	ScrollWindowEx(hWnd, 0, ndy, &rtScroll, NULL, NULL, NULL, SW_INVALIDATE);
	if (ndy > 0){
		RECT rtTop = {0, 0, width, ndy};
		InvalidateRect(hWnd, &rtTop, TRUE);
	}
	if (nOldPage+ndy < (signed)nNewPage){
		RECT rtBottom = {0, nOldPage+ndy, width, nNewPage};
		InvalidateRect(hWnd, &rtBottom, TRUE);
	}
	return;
}

void LevelManager::set_scroll_pos(int pdx, int pdy){
	int width = get_page_h(), height = get_page_v();
	int pos_h = max(0, min(map_cx*32-width, pdx-width/2));
	int pos_v = max(0, min(map_cy*32-height, pdy-height/2));
	set_position_h(pos_h);
	set_position_v(pos_v);
	redraw_default();
	return;
}

bool LevelManager::is_point_in_screen(int pdx, int pdy){
	int pos_h = get_pos_h();
	int pos_v = get_pos_v();
	return pdx >= pos_h && pdx < pos_h+(signed)get_page_h() && pdy >= pos_v && pdy < pos_v+(signed)get_page_v();
}

void LevelManager::clear(){
	sprites.clear();
	bananas.clear();
	cameras.clear();
	return;
}

void LevelManager::reset(){
	load_level();
	return;
}

void LevelManager::load_level(int _level_num){
	level_num = _level_num;
	load_level();
	return;
}


void LevelManager::load_level(){
	SADRS left, right, sdr;
	recent_levels.level_opened(level_num);	// ֒ǉ
	// Xe[W̐ݒ̓ǂݍ
	int level_setting_count = GetPrivateProfileInt("LevelSetting", "Count", 206, RomInfoFileName);
	SADRS adrs_setting = project->read_sadrs(0x3D, 2*level_num);
	if (setting.load_data(project, 0x3D, adrs_setting) == -1){
		MessageBox(project->get_hWnd(), "Xe[Wݒ̓ǂݍ݂Ɏs܂B", "G[",
			MB_OK | MB_ICONEXCLAMATION);
	}else{
		right = adrs_setting + setting.get_size();
		left = GetPrivateProfileInt("LevelSetting", "AvailableEnd", 0x5FEE, RomInfoFileName);
		for (int i=0; i<level_setting_count; i++){
			sdr = project->read_sadrs(0x3D, 2*i);
			if (sdr >= right && sdr < left) left = sdr;
		}
		setting_max_size = left - adrs_setting;
	}
	sprites.load(level_num);	// XvCgf[^̓ǂݍ
	// \̍XV
	redraw_all();
	return;
}

void LevelManager::level_setting(){
	LevelSettingData setting_temp;
	LevelSetting_UndoMethod *undo;
	SADRS adrs_setting;
	LevelSettingDialog dialog(&setting, setting_max_size, level_num);
	int r = dialog.dialog_box(project);
	if (r == IDCANCEL) return;
	setting_temp = setting;
	dialog.get_data(&setting);
	// Undof[^̍쐬EUndoobt@֒ǉ
	undo = new LevelSetting_UndoMethod(this, &setting_temp, &setting, level_num);
	project->undo_buffer.add_method(undo);
	// ROMobt@ɏ߂
	adrs_setting = project->read_sadrs(0x3D, 2*level_num);
	setting.store_data(project, 0x3D, adrs_setting, setting_max_size);
	// \̍XV
	redraw_all();
	return;
}


void LevelManager::redraw(const RECT *rtGlobal){
	RECT rtRedraw = *rtGlobal;
	rtRedraw.right++, rtRedraw.bottom++;
	buffer.select_object(GetStockObject(NULL_PEN));
	buffer.select_object(GetStockObject(LTGRAY_BRUSH));
	buffer.rectangle(0, 0, rtRedraw.right-rtRedraw.left, rtRedraw.bottom-rtRedraw.top);
	tilemap.draw(&buffer, &rtRedraw);
	cameras.draw(&buffer, &rtRedraw);
	bananas.draw(&buffer, &rtRedraw);
	sprites.draw(&buffer, &rtRedraw);

	cameras.draw_f(&buffer, &rtRedraw);
	bananas.draw_f(&buffer, &rtRedraw);
	sprites.draw_f(&buffer, &rtRedraw);
	return;
}

void LevelManager::redraw_default(){
	InvalidateRect(hWnd, NULL, TRUE);
	return;
}

void LevelManager::redraw_all(){
	RECT rt;
	HDC hdc_wnd = GetDC(hWnd);
	tilemap.load(&setting, hdc_wnd);
	ReleaseDC(hWnd, hdc_wnd);
	map_cx = tilemap.get_cx();
	map_cy = tilemap.get_cy();
	map_vert = tilemap.get_vert();
	map_code = tilemap.get_code();
	set_range_h(0, map_cx*32-1);
	set_range_v(0, map_cy*32-1);
	set_position_h(0);
	set_position_v(0);
	GetClientRect(hWnd, &rt);
	set_page_size_h(rt.right);
	set_page_size_v(rt.bottom);
	bananas.load(level_num);	// oiizuf[^̓ǂݍ
	cameras.load(level_num);	// Jzuf[^̓ǂݍ
	redraw_default();
	return;
}

void LevelManager::set_state(bool bEnabled){
	HMENU hmMain = GetMenu(project->get_hWnd());
	UINT uState = (bEnabled ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND;
	EnableMenuItem(GetSubMenu(hmMain, 2), 0, (bEnabled ? MF_ENABLED : MF_GRAYED) | MF_BYPOSITION);
	EnableMenuItem(hmMain, IDM_VIEW_SPRITES,	uState);
	EnableMenuItem(hmMain, IDM_VIEW_BANANAS,	uState);
	EnableMenuItem(hmMain, IDM_VIEW_CAMERAS,	uState);
	EnableMenuItem(hmMain, IDM_LEVEL_SELECT,	uState);
	EnableMenuItem(hmMain, IDM_LEVEL_SETTING,	uState);
	EnableMenuItem(GetSubMenu(hmMain, 3), 1, (bEnabled ? MF_ENABLED : MF_GRAYED) | MF_BYPOSITION);
	ShowWindow(hWnd, bEnabled ? SW_SHOW : SW_HIDE);
	sprites.set_state(bEnabled);
	bananas.set_state(bEnabled);
	cameras.set_state(bEnabled);
	return;
}

void LevelManager::mouseover(LONG x, LONG y){
	int pdx = x + get_pos_h();
	int pdy = y + get_pos_v();
	if (!sprites.mouseover(pdx, pdy)){
		if (!bananas.mouseover(pdx, pdy)){
			if (!cameras.mouseover(pdx, pdy)){
				SetCursor(LoadCursor(NULL, IDC_ARROW));
			}
		}
	}
	return;
}

void LevelManager::clicked(LONG x, LONG y){
	int pdx = x + get_pos_h();
	int pdy = y + get_pos_v();
	SetCapture(hWnd);
	bDragging = false;	// Kʈړď߂ăhbOJn
	mouse_drag_point.x = pdx;
	mouse_drag_point.y = pdy;
	// XvCgɔ邩ǂ
	if (sprites.accepts_click(pdx, pdy)){
		excluse_pages(LEVEL_SPRITES);
		select_page(LEVEL_SPRITES);
		mouse_drag = DRAG_SPRITE;
	}else{
		// oiiɔ邩ǂ
		if (bananas.accepts_click(pdx, pdy)){
			excluse_pages(LEVEL_BANANAS);
			select_page(LEVEL_BANANAS);
			mouse_drag = DRAG_BANANA;
		}else{
			// Jɔ邩ǂ
			if (cameras.accepts_click(pdx, pdy)){
				excluse_pages(LEVEL_CAMERAS);
				select_page(LEVEL_CAMERAS);
				mouse_drag = DRAG_CAMERA;
			}else{
				excluse_pages(LEVEL_UNKNOWN);
				mouse_drag = DRAG_MAP;
			}
		}
	}
	redraw_default();
	return;
}

void LevelManager::unclicked(LONG x, LONG y){
	int pdx = x + get_pos_h();
	int pdy = y + get_pos_v();
	switch (mouse_drag){
		case DRAG_MAP:
			break;
		case DRAG_SPRITE:
			sprites.release_click(bDragging);
			break;
		case DRAG_BANANA:
			bananas.release_click(bDragging);
			break;
		case DRAG_CAMERA:
			cameras.release_click(bDragging);
			break;
	}
	bDragging = false;
	mouse_drag = DRAG_NONE;
	ReleaseCapture();
	redraw_default();
	return;
}

void LevelManager::dragging(LONG x, LONG y){
	int pdx = x + get_pos_h();
	int pdy = y + get_pos_v();
	if (!bDragging){
		int dx = pdx - mouse_drag_point.x;
		int dy = pdy - mouse_drag_point.y;
		if (dx>=-4 && dx<4 && dy>=-4 && dy<4) return;
		bDragging = true;
	}
	switch (mouse_drag){
		case DRAG_MAP:
			break;
		case DRAG_SPRITE:
			sprites.dragging(pdx, pdy);
			break;
		case DRAG_BANANA:
			bananas.dragging(pdx, pdy);
			break;
		case DRAG_CAMERA:
			cameras.dragging(pdx, pdy);
			break;
	}
	redraw_default();
	return;
}

void LevelManager::excluse_pages(LevelEditPage page_selected){
	switch (page_selected){
		case LEVEL_UNKNOWN:
			sprites.release_sel();
			bananas.release_sel();
			cameras.release_sel();
			break;
		case LEVEL_SPRITES:
			bananas.release_sel();
			cameras.release_sel();
			break;
		case LEVEL_BANANAS:
			sprites.release_sel();
			cameras.release_sel();
			break;
		case LEVEL_CAMERAS:
			sprites.release_sel();
			bananas.release_sel();
			break;
	}
	return;
}

void LevelManager::select_page(LevelEditPage page_to_select){
	switch (page_to_select){
		case LEVEL_SPRITES:
			inspector.select_page(0);
			break;
		case LEVEL_BANANAS:
			inspector.select_page(1);
			break;
		case LEVEL_CAMERAS:
			inspector.select_page(2);
			break;
	}
	return;
}


BOOL LevelManager::command(WPARAM wParam, LPARAM lParam){
	if (recent_levels.command(wParam, lParam)) return TRUE;
	if (tilemap.command(wParam, lParam)) return TRUE;
	if (sprites.command(wParam, lParam)) return TRUE;
	if (bananas.command(wParam, lParam)) return TRUE;
	if (cameras.command(wParam, lParam)) return TRUE;
	switch (LOWORD(wParam)){
		case IDM_LEVEL_SELECT:
			{
				LevelSelectDialog level_sel;
				int r = level_sel.dialog_box(project, level_num);
				if (r == -1) return TRUE;
				level_num = r;
				load_level();
			}
			return TRUE;
		/*case IDM_LEVEL_SETTING:
			level_setting();
			return TRUE;*/
	}
	return FALSE;
}














