
///////////////////////////////////////////////////////////////////////////////
//
//	level_setting.cpp
//

#include <stdio.h>
#include "id.h"
#include "level_select.h"
#include "project.h"



int LevelSettingData::get_size() const{
	int size = 18;
	if (type == 0x0004 || type == 0x0005){
		return 2;
	}
	if (type == 0x0001) size++;
	size += start_count * 5;
	size += jump_count * 2;
	size += uparam_count * 4;
	return size;
}

int LevelSettingData::load_data(const ProjectManager *project, BANK bank, SADRS adrs){
	int i=0;

	type = project->read_word(bank, adrs+i);			i += 2;
	if (type == 0x0002){
		i += 2;
	}
	i++;

	style		= project->read_sadrs(bank, adrs+i);	i += 2;
	map_index	= project->read_byte(bank, adrs+i);		i++;

	/*
	type = project->read_word(bank, adrs+i);			i += 2;
	if (type == 0x0004 || type == 0x0005) return i;
	if (type == 0x0001){
		bonus_type = project->read_byte(bank, adrs+i);	i++;
	}
	style		= project->read_sadrs(bank, adrs+i);	i += 2;
	map_index	= project->read_byte(bank, adrs+i);		i++;
	param1		= project->read_byte(bank, adrs+i);		i++;
	param2		= project->read_byte(bank, adrs+i);		i++;
	param3		= project->read_byte(bank, adrs+i);		i++;
	param4		= project->read_byte(bank, adrs+i);		i++;
	param5		= project->read_byte(bank, adrs+i);		i++;
	param6		= project->read_byte(bank, adrs+i);		i++;
	start_count = 0;
	while (1){
		BYTE byte = project->read_byte(bank, adrs+i);	i++;
		if (byte == 0xFF) break;
		if (start_count == START_MAX){
			return -1;
		}
		starts[start_count].param	= byte;
		starts[start_count].pdx		= (signed)project->read_word(bank, adrs+i) - 256;	i+=2;
		starts[start_count].pdy		= (signed)project->read_word(bank, adrs+i) - 256;	i+=2;
		start_count++;
	}
	jump_count = 0;
	while (1){
		WORD word = project->read_word(bank, adrs+i);	i+=2;
		if (word == 0xFFFF) break;
		if (jump_count == JUMP_MAX){
			return -1;
		}
		jumps[jump_count] = word;
		jump_count++;
	}
	i += 2;		// 0000΂
	uparam_count = 0;
	while (1){
		WORD word = project->read_word(bank, adrs+i);	i+=2;
		if (word == 0xFFFF) break;
		if (uparam_count == UPARAM_MAX){
			return -1;
		}
		uparams[uparam_count].pl = word;
		uparams[uparam_count].ph = project->read_word(bank, adrs+i);	i+=2;
		uparam_count++;
	}
	//*/
	return i;
}

int LevelSettingData::store_data(ProjectManager *project, BANK bank, SADRS adrs, int size_max) const{
	/*
	int i=0;
	if (get_size() > size_max) return -1;
	project->write_word(type, bank, adrs+i);			i+=2;
	if (type == 0x0004 || type == 0x0005) return i;
	if (type == 0x0001){
		project->write_byte(bonus_type, bank, adrs+i);	i++;
	}
	project->write_word(style,		bank, adrs+i);		i+=2;
	project->write_byte(map_index,	bank, adrs+i);		i++;
	project->write_byte(param1,		bank, adrs+i);		i++;
	project->write_byte(param2,		bank, adrs+i);		i++;
	project->write_byte(param3,		bank, adrs+i);		i++;
	project->write_byte(param4,		bank, adrs+i);		i++;
	project->write_byte(param5,		bank, adrs+i);		i++;
	project->write_byte(param6,		bank, adrs+i);		i++;
	for (int k=0; k<start_count; k++){
		project->write_byte(starts[k].param, bank, adrs+i);		i++;
		project->write_word(starts[k].pdx+256, bank, adrs+i);	i+=2;
		project->write_word(starts[k].pdy+256, bank, adrs+i);	i+=2;
	}
	project->write_byte(0xFF, bank, adrs+i);			i++;
	for (int k=0; k<jump_count; k++){
		project->write_word(jumps[k], bank, adrs+i);	i+=2;
	}
	project->write_word(0xFFFF, bank, adrs+i);			i+=2;
	project->write_word(0x0000, bank, adrs+i);			i+=2;
	for (int k=0; k<uparam_count; k++){
		project->write_word(uparams[k].pl, bank, adrs+i);		i+=2;
		project->write_word(uparams[k].ph, bank, adrs+i);		i+=2;
	}
	project->write_word(0xFFFF, bank, adrs+i);
	return i;
	*/
	return 0;
}


//=============================================================================


static struct{
	int value;
	LPCSTR name;
} level_type[7] = {
	{0, "0: ʏ"},
	{1, "1: {[iX"},
	{2, "2: "},
	{3, "3: {X"},
	{4, "4: gp̃Xe[W"},
	{5, "5: gp̃Xe[W"},
	{6, "6: Wv"}
};
static const int level_type_count = 7;

static struct {
	int value;
	LPCSTR name;
} bonus_type[4] = {
	{0, "0: Ȃ"},
	{1, "1: ݂Ȃ"},
	{2, "2: ق߂"},
	{3, "3: RC"}
};
static const int bonus_type_count = 4;



LPCSTR LevelSettingDialog::get_template() const{
	return "LEVEL_SETTING";
}


LevelSettingDialog::LevelSettingDialog(const LevelSettingData *data_src, int _max_size, int _level_num){
	data = *data_src;
	max_size = _max_size;
	level_num = _level_num;
	if (data.type == 0x0004 || data.type == 0x0005){
		data.start_count = 0;
		data.jump_count = 0;
		data.uparam_count = 0;
	}
}
LevelSettingDialog::~LevelSettingDialog(){
}

void LevelSettingDialog::get_data(LevelSettingData *data_dst){
	*data_dst = data;
	return;
}


BOOL LevelSettingDialog::Proc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam){
	switch (msg){
		case WM_INITDIALOG:
			init_dialog();
			return TRUE;
		case WM_COMMAND:
			switch (LOWORD(wParam)){
				case IDOK:
					if (!end_setting()) return TRUE;
					EndDialog(hWnd, IDOK);
					return TRUE;
				case IDCANCEL:
					EndDialog(hWnd, IDCANCEL);
					return TRUE;
				case IDC_BUTTON9:
					select_level();
					return TRUE;
				case IDC_BUTTON3:
					add_start();
					return TRUE;
				case IDC_BUTTON4:
					delete_start();
					return TRUE;
				case IDC_BUTTON5:
					add_jump();
					return TRUE;
				case IDC_BUTTON6:
					delete_jump();
					return TRUE;
				case IDC_BUTTON7:
					add_uparam();
					return TRUE;
				case IDC_BUTTON8:
					delete_uparam();
					return TRUE;
				case IDC_COMBO0:
					if (HIWORD(wParam) == CBN_SELCHANGE){
						level_type_changed((WORD)SendMessage(GetDlgItem(hWnd, IDC_COMBO0), CB_GETCURSEL, 0, 0));
						return TRUE;
					}
					return FALSE;
				case IDC_LIST0:
					if (HIWORD(wParam) == LBN_SELCHANGE){
						start_selchanged(SendMessage(GetDlgItem(hWnd, IDC_LIST0), LB_GETCURSEL, 0, 0));
						return TRUE;
					}
					return FALSE;
				case IDC_LIST1:
					if (HIWORD(wParam) == LBN_SELCHANGE){
						jump_selchanged(SendMessage(GetDlgItem(hWnd, IDC_LIST1), LB_GETCURSEL, 0, 0));
						return TRUE;
					}
					return FALSE;
				case IDC_LIST2:
					if (HIWORD(wParam) == LBN_SELCHANGE){
						uparam_selchanged(SendMessage(GetDlgItem(hWnd, IDC_LIST2), LB_GETCURSEL, 0, 0));
						return TRUE;
					}
					return FALSE;
				case IDC_EDIT9:
				case IDC_EDITA:
				case IDC_EDITB:
					if (HIWORD(wParam) == EN_CHANGE){
						if (start_active) apply_start();
						return TRUE;
					}
					return FALSE;
				case IDC_EDITE:
					if (HIWORD(wParam) == EN_CHANGE){
						if (jump_active) apply_jump();
						return TRUE;
					}
					return FALSE;
				case IDC_EDITC:
				case IDC_EDITD:
					if (HIWORD(wParam) == EN_CHANGE){
						if (uparam_active) apply_uparam();
						return TRUE;
					}
					return FALSE;
			}
			return FALSE;
	}
	return FALSE;
}


void LevelSettingDialog::init_dialog(){
	char str[64], str2[64];
	bool b1;
	int n;
	
	HWND hwLevelTypeList = GetDlgItem(hWnd, IDC_COMBO0);
	HWND hwBonusTypeList = GetDlgItem(hWnd, IDC_COMBO1);
	HWND hwLevelStyleList = GetDlgItem(hWnd, IDC_COMBO2);
	const int level_style_count = GetPrivateProfileInt("Styles", "Count", 52, RomInfoFileName);
	const int level_style_begin = GetPrivateProfileInt("Styles", "Begin", 0x479E, RomInfoFileName);

	HWND hwStartList	= GetDlgItem(hWnd, IDC_LIST0);
	HWND hwJumpList		= GetDlgItem(hWnd, IDC_LIST1);
	HWND hwUParamList	= GetDlgItem(hWnd, IDC_LIST2);

	//	
	//-------------------------------------------------------------------------
	for (int i=0; i<level_type_count; i++){
		SendMessage(hwLevelTypeList, CB_INSERTSTRING, (WPARAM)i, (LPARAM)level_type[i].name);
	}
	for (int i=0; i<bonus_type_count; i++){
		SendMessage(hwBonusTypeList, CB_INSERTSTRING, (WPARAM)i, (LPARAM)bonus_type[i].name);
	}
	for (int i=0, j=level_style_begin; i<level_style_count; i++, j+=21){
		sprintf_s(str, sizeof str, "%04X", j);
		GetPrivateProfileString("StyleName", str, "?", str2, sizeof str2, RomInfoFileName);
		sprintf_s(str, sizeof str, "%04X: %s", j, str2);
		SendMessage(hwLevelStyleList, CB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
	}

	//	̔f
	//-------------------------------------------------------------------------
	start_active = true;
	jump_active = true;
	uparam_active = true;
	// LvV
	sprintf_s(str, sizeof str, "%02X", level_num);
	GetPrivateProfileString("LevelName", str, "?", str2, sizeof str2, RomInfoFileName);
	sprintf_s(str, sizeof str, "Xe[W̐ݒ[%02X: %s]", level_num, str2);
	set_text(str);
	// Xe[W̎
	b1 = true;
	for (int i=0; i<level_type_count; i++){
		if (data.type == level_type[i].value){
			SendMessage(hwLevelTypeList, CB_SETCURSEL, (WPARAM)i, (LPARAM)0);
			b1 = false;
			break;
		}
	}
	if (b1){
		MessageBox(hWnd, "Xe[W̎ނzOłB", "x", MB_OK | MB_ICONEXCLAMATION);
	}
	level_type_changed(data.type);
	if (data.type != 0x0004 && data.type != 0x0005){
		if (data.type == 0x0001){
			b1 = true;
			for (int i=0; i<bonus_type_count; i++){
				if (data.bonus_type == bonus_type[i].value){
					SendMessage(hwBonusTypeList, CB_SETCURSEL, (WPARAM)i, (LPARAM)0);
					b1 = false;
					break;
				}
			}
			if (b1){
				MessageBox(hWnd, "{[iX̎ނzOłB", "x", MB_OK | MB_ICONEXCLAMATION);
			}
		}
		sprintf_s(str, sizeof str, "%02X", data.map_index);
		SetWindowText(GetDlgItem(hWnd, IDC_EDIT2), str);
		n = (data.style - level_style_begin) / 21;
		SendMessage(hwLevelStyleList, CB_SETCURSEL, (WPARAM)n, (LPARAM)0);
		sprintf_s(str, sizeof str, "%02X", data.param1);
		SetWindowText(GetDlgItem(hWnd, IDC_EDIT3), str);
		sprintf_s(str, sizeof str, "%02X", data.param2);
		SetWindowText(GetDlgItem(hWnd, IDC_EDIT4), str);
		sprintf_s(str, sizeof str, "%02X", data.param3);
		SetWindowText(GetDlgItem(hWnd, IDC_EDIT5), str);
		sprintf_s(str, sizeof str, "%02X", data.param4);
		SetWindowText(GetDlgItem(hWnd, IDC_EDIT6), str);
		sprintf_s(str, sizeof str, "%02X", data.param5);
		SetWindowText(GetDlgItem(hWnd, IDC_EDIT7), str);
		sprintf_s(str, sizeof str, "%02X", data.param6);
		SetWindowText(GetDlgItem(hWnd, IDC_EDIT8), str);
		for (int i=0; i<data.start_count; i++){
			get_text_start(str, sizeof str, i, data.starts[i].param, data.starts[i].pdx, data.starts[i].pdy);
			SendMessage(hwStartList, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
		}
		start_selchanged(-1);
		for (int i=0; i<data.jump_count; i++){
			get_text_jump(str, sizeof str, i, data.jumps[i]);
			SendMessage(hwJumpList, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
		}
		jump_selchanged(-1);
		for (int i=0; i<data.uparam_count; i++){
			get_text_uparam(str, sizeof str, i, data.uparams[i].pl, data.uparams[i].ph);
			SendMessage(hwUParamList, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
		}
		uparam_selchanged(-1);
	}
	return;
}


void LevelSettingDialog::level_type_changed(WORD new_value){
	BOOL bEnabled_setting = (new_value != 0x0004) && (new_value != 0x0005);
	BOOL bEnabled_bonus   = (new_value == 0x0001);
	EnableWindow(GetDlgItem(hWnd, IDC_COMBO1), bEnabled_bonus);
	EnableWindow(GetDlgItem(hWnd, IDC_COMBO2), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT2), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT3), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT4), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT5), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT6), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT7), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT8), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_LIST0), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT9), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDITA), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDITB), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON3), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON4), bEnabled_setting
		&& (SendMessage(GetDlgItem(hWnd, IDC_LIST0), LB_GETCURSEL, 0, 0) != -1));
	EnableWindow(GetDlgItem(hWnd, IDC_LIST1), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDITE), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDITF), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON9), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON5), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON6), bEnabled_setting
		&& (SendMessage(GetDlgItem(hWnd, IDC_LIST1), LB_GETCURSEL, 0, 0) != -1));
	EnableWindow(GetDlgItem(hWnd, IDC_LIST2), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDITC), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_EDITD), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON7), bEnabled_setting);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON8), bEnabled_setting
		&& (SendMessage(GetDlgItem(hWnd, IDC_LIST2), LB_GETCURSEL, 0, 0) != -1));
	return;
}

void LevelSettingDialog::start_selchanged(int start_num){
	char str[64];
	BOOL bEnabled = start_num != -1;
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON4), bEnabled);
	if (bEnabled){
		start_active = false;
		sprintf_s(str, sizeof str, "%02X", data.starts[start_num].param);
		SetDlgItemText(hWnd, IDC_EDIT9, str);
		sprintf_s(str, sizeof str, "%d", data.starts[start_num].pdx);
		SetDlgItemText(hWnd, IDC_EDITA, str);
		sprintf_s(str, sizeof str, "%d", data.starts[start_num].pdy);
		SetDlgItemText(hWnd, IDC_EDITB, str);
		start_active = true;
	}
	return;
}
void LevelSettingDialog::jump_selchanged(int jump_num){
	char str[64], str2[64];
	BOOL bEnabled = jump_num != -1;
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON6), bEnabled);
	if (bEnabled){
		jump_active = false;
		sprintf_s(str, sizeof str, "%04X", data.jumps[jump_num]);
		SetDlgItemText(hWnd, IDC_EDITE, str);
		sprintf_s(str2, sizeof str2, "%02X", data.jumps[jump_num]);
		GetPrivateProfileString("LevelName", str2, "?", str, sizeof str, RomInfoFileName);
		SetDlgItemText(hWnd, IDC_EDITF, str);
		jump_active = true;
	}
	return;
}
void LevelSettingDialog::uparam_selchanged(int uparam_num){
	char str[64];
	BOOL bEnabled = uparam_num != -1;
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON8), bEnabled);
	if (bEnabled){
		uparam_active = false;
		sprintf_s(str, sizeof str, "%04X", data.uparams[uparam_num].pl);
		SetDlgItemText(hWnd, IDC_EDITC, str);
		sprintf_s(str, sizeof str, "%04X", data.uparams[uparam_num].ph);
		SetDlgItemText(hWnd, IDC_EDITD, str);
		uparam_active = true;
	}
	return;
}

void LevelSettingDialog::get_text_start(char *dst, int nSize, int index, BYTE param, int pdx, int pdy){
	sprintf_s(dst, nSize, "#%d: [%02X] (%d, %d)", index, param, pdx, pdy);
	return;
}
void LevelSettingDialog::get_text_jump(char *dst, int nSize, int index, WORD jump){
	char str[64];
	sprintf_s(dst, nSize, "%02X", jump);
	GetPrivateProfileString("LevelName", dst, "?", str, sizeof str, RomInfoFileName);
	sprintf_s(dst, nSize, "#%d: %04X[%s]", index, jump, str);
	return;
}
void LevelSettingDialog::get_text_uparam(char *dst, int nSize, int index, WORD pl, WORD ph){
	sprintf_s(dst, nSize, "#%d: (%04X, %04X)", index, pl, ph);
	return;
}

void LevelSettingDialog::apply_start(){
	char str[64];
	int value;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	int isel = SendMessage(hwList, LB_GETCURSEL, 0, 0);
	if (isel == -1) return;
	GetDlgItemText(hWnd, IDC_EDIT9, str, sizeof str);
	if (sscanf_s(str, "%x", &value) < 1) value = 0;
	data.starts[isel].param = (BYTE)value;
	GetDlgItemText(hWnd, IDC_EDITA, str, sizeof str);
	if (sscanf_s(str, "%d", &value) < 1) value = 0;
	data.starts[isel].pdx = value;
	GetDlgItemText(hWnd, IDC_EDITB, str, sizeof str);
	if (sscanf_s(str, "%d", &value) < 1) value = 0;
	data.starts[isel].pdy = value;
	get_text_start(str, sizeof str, isel, data.starts[isel].param, data.starts[isel].pdx, data.starts[isel].pdy);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	SendMessage(hwList, LB_DELETESTRING, (WPARAM)isel, 0);
	SendMessage(hwList, LB_INSERTSTRING, (WPARAM)isel, (LPARAM)str);
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	return;
}

void LevelSettingDialog::apply_jump(){
	char str[64], str2[64];
	int value;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST1);
	int isel = SendMessage(hwList, LB_GETCURSEL, 0, 0);
	GetDlgItemText(hWnd, IDC_EDITE, str, sizeof str);
	if (sscanf_s(str, "%x", &value) < 1) value = 0;
	{
		sprintf_s(str2, sizeof str2, "%02X", (WORD)value);
		GetPrivateProfileString("LevelName", str2, "?", str, sizeof str, RomInfoFileName);
		SetDlgItemText(hWnd, IDC_EDITF, str);
	}
	if (isel == -1) return;
	data.jumps[isel] = (WORD)value;
	get_text_jump(str, sizeof str, isel, data.jumps[isel]);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	SendMessage(hwList, LB_DELETESTRING, (WPARAM)isel, 0);
	SendMessage(hwList, LB_INSERTSTRING, (WPARAM)isel, (LPARAM)str);
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	return;
}

void LevelSettingDialog::apply_uparam(){
	char str[64];
	int value;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST2);
	int isel = SendMessage(hwList, LB_GETCURSEL, 0, 0);
	if (isel == -1) return;
	GetDlgItemText(hWnd, IDC_EDITC, str, sizeof str);
	if (sscanf_s(str, "%x", &value) < 1) value = 0;
	data.uparams[isel].pl = (WORD)value;
	GetDlgItemText(hWnd, IDC_EDITD, str, sizeof str);
	if (sscanf_s(str, "%x", &value) < 1) value = 0;
	data.uparams[isel].ph = (WORD)value;
	get_text_uparam(str, sizeof str, isel, data.uparams[isel].pl, data.uparams[isel].ph);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	SendMessage(hwList, LB_DELETESTRING, (WPARAM)isel, 0);
	SendMessage(hwList, LB_INSERTSTRING, (WPARAM)isel, (LPARAM)str);
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	return;
}

void LevelSettingDialog::select_level(){
	char str[64];
	int value;
	LevelSelectDialog dialog;
	GetDlgItemText(hWnd, IDC_EDITE, str, sizeof str);
	if (sscanf_s(str, "%x", &value) < 1) value = 0;
	int r = dialog.dialog_box(this, (LPARAM)value);
	if (r == -1) return;
	sprintf_s(str, sizeof str, "%04X", r);
	SetDlgItemText(hWnd, IDC_EDITE, str);
	return;
}

void LevelSettingDialog::add_start(){
	char str[64];
	int value;
	int isel;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	if (data.start_count + 1 == LevelSettingData::START_MAX){
		MessageBox(hWnd, "ȏX^[gn_ǉł܂B", "G[", MB_OK | MB_ICONEXCLAMATION);
		return;
	}
	isel = SendMessage(hwList, LB_GETCURSEL, 0, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	if (isel == -1) isel = data.start_count;
	for (int i=data.start_count-1; i>=isel; i--){
		data.starts[i+1] = data.starts[i];
		SendMessage(hwList, LB_DELETESTRING, (WPARAM)i, 0);
	}
	data.start_count++;
	GetDlgItemText(hWnd, IDC_EDIT9, str, sizeof str);
	if (sscanf_s(str, "%x", &value) < 1) value = 0;
	data.starts[isel].param = value;
	GetDlgItemText(hWnd, IDC_EDITA, str, sizeof str);
	if (sscanf_s(str, "%d", &value) < 1) value = 0;
	data.starts[isel].pdx = value;
	GetDlgItemText(hWnd, IDC_EDITB, str, sizeof str);
	if (sscanf_s(str, "%d", &value) < 1) value = 0;
	data.starts[isel].pdy = value;
	for (int i=isel; i<data.start_count; i++){
		get_text_start(str, sizeof str, i, data.starts[i].param, data.starts[i].pdx, data.starts[i].pdy);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
	}
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	if (isel+1 == data.start_count){
		start_selchanged(isel);
	}
	return;
}

void LevelSettingDialog::add_jump(){
	char str[64];
	int value;
	int isel;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST1);
	if (data.jump_count + 1 == LevelSettingData::JUMP_MAX){
		MessageBox(hWnd, "ȏړXe[Wǉł܂B", "G[", MB_OK | MB_ICONEXCLAMATION);
		return;
	}
	isel = SendMessage(hwList, LB_GETCURSEL, 0, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	if (isel == -1) isel = data.jump_count;
	for (int i=data.jump_count-1; i>=isel; i--){
		data.jumps[i+1] = data.jumps[i];
		SendMessage(hwList, LB_DELETESTRING, (WPARAM)i, 0);
	}
	data.jump_count++;
	GetDlgItemText(hWnd, IDC_EDITE, str, sizeof str);
	if (sscanf_s(str, "%x", &value) < 1) value = 0;
	data.jumps[isel] = value;
	for (int i=isel; i<data.jump_count; i++){
		get_text_jump(str, sizeof str, i, data.jumps[i]);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
	}
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	if (isel+1 == data.jump_count){
		jump_selchanged(isel);
	}
	return;
}

void LevelSettingDialog::add_uparam(){
	char str[64];
	int value;
	int isel;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST2);
	if (data.uparam_count + 1 == LevelSettingData::UPARAM_MAX){
		MessageBox(hWnd, "ȏp[^ǉł܂B", "G[", MB_OK | MB_ICONEXCLAMATION);
		return;
	}
	isel = SendMessage(hwList, LB_GETCURSEL, 0, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	if (isel == -1) isel = data.uparam_count;
	for (int i=data.uparam_count-1; i>=isel; i--){
		data.uparams[i+1] = data.uparams[i];
		SendMessage(hwList, LB_DELETESTRING, (WPARAM)i, 0);
	}
	data.uparam_count++;
	GetDlgItemText(hWnd, IDC_EDITC, str, sizeof str);
	if (sscanf_s(str, "%x", &value) < 1) value = 0;
	data.uparams[isel].pl = value;
	GetDlgItemText(hWnd, IDC_EDITD, str, sizeof str);
	if (sscanf_s(str, "%x", &value) < 1) value = 0;
	data.uparams[isel].ph = value;
	for (int i=isel; i<data.uparam_count; i++){
		get_text_uparam(str, sizeof str, i, data.uparams[i].pl, data.uparams[i].ph);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
	}
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	if (isel+1 == data.uparam_count){
		uparam_selchanged(isel);
	}
	return;
}

void LevelSettingDialog::delete_start(){
	char str[64];
	int isel;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	isel = SendMessage(hwList, LB_GETCURSEL, 0, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	if (isel == -1) return;
	for (int i=isel; i<data.start_count; ){
		SendMessage(hwList ,LB_DELETESTRING, (WPARAM)isel, 0);
		if (++i < data.start_count){
			data.starts[i-1] = data.starts[i];
		}
	}
	data.start_count--;
	for (int i=isel; i<data.start_count; i++){
		get_text_start(str, sizeof str, i, data.starts[i].param, data.starts[i].pdx, data.starts[i].pdy);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
	}
	if (isel == data.start_count) isel = data.start_count-1;
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	start_selchanged(isel);
	return;
}

void LevelSettingDialog::delete_jump(){
	char str[64];
	int isel;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST1);
	isel = SendMessage(hwList, LB_GETCURSEL, 0, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	if (isel == -1) return;
	for (int i=isel; i<data.jump_count; ){
		SendMessage(hwList ,LB_DELETESTRING, (WPARAM)isel, 0);
		if (++i < data.jump_count){
			data.jumps[i-1] = data.jumps[i];
		}
	}
	data.jump_count--;
	for (int i=isel; i<data.jump_count; i++){
		get_text_jump(str, sizeof str, i, data.jumps[i]);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
	}
	if (isel == data.jump_count) isel = data.jump_count-1;
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	jump_selchanged(isel);
	return;
}

void LevelSettingDialog::delete_uparam(){
	char str[64];
	int isel;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST2);
	isel = SendMessage(hwList, LB_GETCURSEL, 0, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	if (isel == -1) return;
	for (int i=isel; i<data.uparam_count; ){
		SendMessage(hwList ,LB_DELETESTRING, (WPARAM)isel, 0);
		if (++i < data.uparam_count){
			data.uparams[i-1] = data.uparams[i];
		}
	}
	data.uparam_count--;
	for (int i=isel; i<data.uparam_count; i++){
		get_text_uparam(str, sizeof str, i, data.uparams[i].pl, data.uparams[i].ph);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)i, (LPARAM)str);
	}
	if (isel == data.uparam_count) isel = data.uparam_count-1;
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)isel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	uparam_selchanged(isel);
	return;
}

bool LevelSettingDialog::end_setting(){
	char str[256];
	int isel, value, size;
	isel = SendMessage(GetDlgItem(hWnd, IDC_COMBO0), CB_GETCURSEL, 0, 0);
	if (isel < 0 || isel >= level_type_count){
		MessageBox(hWnd, "Xe[W̎ނsłB", "G[", MB_OK | MB_ICONEXCLAMATION);
		return false;
	}
	data.type = level_type[isel].value;
	if (data.type != 0x0004 && data.type != 0x0005){
		if (data.type == 0x0001){
			isel = SendMessage(GetDlgItem(hWnd, IDC_COMBO1), CB_GETCURSEL, 0, 0);
			if (isel < 0 || isel >= bonus_type_count){
				MessageBox(hWnd, "{[iX̎ނsłB", "G[", MB_OK | MB_ICONEXCLAMATION);
				return false;
			}
			data.bonus_type = bonus_type[isel].value;
		}
		GetDlgItemText(hWnd, IDC_EDIT2, str, sizeof str);
		if (sscanf_s(str, "%x", &value) < 1) value = 0;
		data.map_index = value;
		const int level_style_count = GetPrivateProfileInt("Styles", "Count", 52, RomInfoFileName);
		const int level_style_begin = GetPrivateProfileInt("Styles", "Begin", 0x479E, RomInfoFileName);
		isel = SendMessage(GetDlgItem(hWnd, IDC_COMBO2), CB_GETCURSEL, 0, 0);
		if (isel < 0 || isel >= level_style_count){
			MessageBox(hWnd, "X^C̒lsłB", "G[", MB_OK | MB_ICONEXCLAMATION);
			return false;
		}
		data.style = level_style_begin + isel*21;
		GetDlgItemText(hWnd, IDC_EDIT3, str, sizeof str);
		if (sscanf_s(str, "%x", &value) < 1) value = 0;
		data.param1 = value;
		GetDlgItemText(hWnd, IDC_EDIT4, str, sizeof str);
		if (sscanf_s(str, "%x", &value) < 1) value = 0;
		data.param2 = value;
		GetDlgItemText(hWnd, IDC_EDIT5, str, sizeof str);
		if (sscanf_s(str, "%x", &value) < 1) value = 0;
		data.param3 = value;
		GetDlgItemText(hWnd, IDC_EDIT6, str, sizeof str);
		if (sscanf_s(str, "%x", &value) < 1) value = 0;
		data.param4 = value;
		GetDlgItemText(hWnd, IDC_EDIT7, str, sizeof str);
		if (sscanf_s(str, "%x", &value) < 1) value = 0;
		data.param5 = value;
		GetDlgItemText(hWnd, IDC_EDIT8, str, sizeof str);
		if (sscanf_s(str, "%x", &value) < 1) value = 0;
		data.param6 = value;
	}
	size = data.get_size();
	if (size > max_size){
		sprintf_s(str, sizeof str, "f[^̃TCY傫܂B\r\n"
			"݂̃TCY: %d\r\nő̃TCY: %d", size, max_size);
		MessageBox(hWnd, str, "G[", MB_OK | MB_ICONEXCLAMATION);
		return false;
	}
	return true;
}





