
///////////////////////////////////////////////////////////////////////////////
//
//	project.cpp
//

#include <stdio.h>
#include <shlwapi.h>
#include "id.h"
#include "project.h"

#pragma warning(disable: 4355)


LPCSTR ProjectManager::pAppNameMainWindow = "MainWindow";
LPCSTR ProjectManager::pClassName = "sdk3ed_main";

BOOL ProjectManager::Init(){
	if (!register_class_ex(
		CS_HREDRAW | CS_VREDRAW,
		(HICON)LoadIcon(hInst, MAKEINTRESOURCE(IDI_SDK3ED)),
		(HICON)LoadIcon(hInst, MAKEINTRESOURCE(IDI_SDK3ED)),
		LoadCursor(NULL, IDC_ARROW),
		(HBRUSH)(COLOR_APPWORKSPACE + 1),
		pClassName,
		"SDK3ED_MENU"))
	{
		return FALSE;
	}
	if (!LevelManager::Init()) return FALSE;
	return TRUE;
}

ProjectManager::ProjectManager()
: undo_buffer(), levels(this), recent_files(this), test_play(this)
{
	bEnabled = FALSE;
	nChangeStep = 0;
	bChangedInrev = false;
	buf = NULL;
	header = new BYTE[ROM_HEADER_SIZE];
}


ProjectManager::~ProjectManager(){
	if (buf != NULL) delete[] buf;
	delete[] header;
}


BOOL ProjectManager::create(LPCSTR cmdline, int nCmdShow){

	int x, y, w, h, m;
	x = (signed)GetPrivateProfileInt(pAppNameMainWindow, "Pos_X", CW_USEDEFAULT, ConfigFileName);
	y = (signed)GetPrivateProfileInt(pAppNameMainWindow, "Pos_Y", CW_USEDEFAULT, ConfigFileName);
	w = (signed)GetPrivateProfileInt(pAppNameMainWindow, "Width", CW_USEDEFAULT, ConfigFileName);
	h = (signed)GetPrivateProfileInt(pAppNameMainWindow, "Height", CW_USEDEFAULT, ConfigFileName);
	m = (signed)GetPrivateProfileInt(pAppNameMainWindow, "Maximized", 0, ConfigFileName);
	if ((x==CW_USEDEFAULT)||(y==CW_USEDEFAULT)||(w==CW_USEDEFAULT)||(h==CW_USEDEFAULT)){
		x = y = w = h = CW_USEDEFAULT;
	}

	if (!create_window_ex(pClassName, "SDK3 Editor",
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
		0,
		x, y, w, h,
		NULL))
	{
		return FALSE;
	}
	DragAcceptFiles(hWnd, TRUE);
	set_state(false);
	open_cmdline(cmdline);
	ShowWindow(hWnd, m ? SW_SHOWMAXIMIZED : nCmdShow);
	UpdateWindow(hWnd);
	return TRUE;
}

void ProjectManager::save_window_pos(){
	WINDOWPLACEMENT wp;
	GetWindowPlacement(hWnd, &wp);
	char s[16];
	sprintf_s(s, sizeof s, "%d", wp.rcNormalPosition.left);
	WritePrivateProfileString(pAppNameMainWindow, "Pos_X", s, ConfigFileName);
	sprintf_s(s, sizeof s, "%d", wp.rcNormalPosition.top);
	WritePrivateProfileString(pAppNameMainWindow, "Pos_Y", s, ConfigFileName);
	sprintf_s(s, sizeof s, "%d", wp.rcNormalPosition.right - wp.rcNormalPosition.left);
	WritePrivateProfileString(pAppNameMainWindow, "Width", s, ConfigFileName);
	sprintf_s(s, sizeof s, "%d", wp.rcNormalPosition.bottom - wp.rcNormalPosition.top);
	WritePrivateProfileString(pAppNameMainWindow, "Height", s, ConfigFileName);
	sprintf_s(s, sizeof s, "%d", wp.showCmd == SW_SHOWMAXIMIZED);
	WritePrivateProfileString(pAppNameMainWindow, "Maximized", s, ConfigFileName);
	return;
}

void ProjectManager::set_state(bool _bEnabled){
	bEnabled = _bEnabled;
	HMENU hmMain = GetMenu(hWnd);
	UINT uState = (bEnabled ? MF_ENABLED : MF_GRAYED) | MF_BYCOMMAND;
	if (bEnabled){
		char title[256];
		sprintf_s(title, sizeof title, "%s - SDK3 Editor", filetitle);
		this->set_text(title);
	}else{
		this->set_text("SDK3 Editor");
	}
	EnableMenuItem(hmMain, IDM_FILE_SAVE,		uState);
	EnableMenuItem(hmMain, IDM_FILE_SAVEAS,		uState);
	EnableMenuItem(hmMain, IDM_FILE_RELOAD,		uState);
	EnableMenuItem(hmMain, IDM_FILE_CLOSE,		uState);
	EnableMenuItem(hmMain, IDM_FILE_TESTPLAY,	uState);
	EnableMenuItem(hmMain, IDM_EDIT_UNDO,		uState);
	EnableMenuItem(hmMain, IDM_EDIT_REDO,		uState);
	levels.set_state(bEnabled);
	return;
}

LRESULT ProjectManager::Proc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam){
	switch (msg){
		case WM_CREATE:
			if (!recent_files.create()) return -1;
			if (!undo_buffer.create(this)) return -1;
			if (!levels.create()) return -1;
			return 0;
		case WM_ACTIVATE:
			//if (LOWORD(wParam) == WA_INACTIVE && (HWND)lParam == levels.inspector.get_hWnd()) break;
			levels.inspector.nc_activate(wParam);
			break;
		case WM_SIZE:
			levels.move(0, 0, LOWORD(lParam), HIWORD(lParam), TRUE);
			return 0;
		case WM_ENABLE:
			EnableWindow(levels.inspector.get_hWnd(), wParam ? TRUE : FALSE);
			return 0;
		case WM_COMMAND:
			if (command(wParam, lParam)) return 0;
			break;
		case WM_SETFOCUS:
			SetFocus(levels.get_hWnd());
			return 0;
		case WM_DROPFILES:
			open_dropped_files((HDROP)wParam);
			return 0;
		case WM_CLOSE:
			quit();
			return 0;
		case WM_DESTROY:
			recent_files.save_data();
			save_window_pos();
			PostQuitMessage(0);
			return 0;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}

void ProjectManager::nc_activate(WPARAM wParam){
	SetWindowPos(hWnd, HWND_TOP, 0, 0, 0, 0,
		SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
	//DefWindowProc(hWnd, WM_NCACTIVATE, wParam, 0);
	return;
}

void ProjectManager::changed_normal(){
	if (bChangedInrev) return;
	if (nChangeStep >= 0){
		nChangeStep++;			// Xebv(t)
	}else{
		bChangedInrev = true;	// st
	}
	return;
}

void ProjectManager::changed_undo(){
	if (bChangedInrev) return;
	nChangeStep--;
	return;
}

void ProjectManager::changed_redo(){
	if (bChangedInrev) return;
	nChangeStep++;
	return;
}

BOOL ProjectManager::command(WPARAM wParam, LPARAM lParam){
	if (recent_files.command(wParam, lParam)) return TRUE;
	if (test_play.command(wParam, lParam)) return TRUE;
	if (bEnabled){
		if (undo_buffer.command(wParam, lParam)) return TRUE;
		if (levels.command(wParam, lParam)) return TRUE;
	}
	switch (LOWORD(wParam)){
		case IDM_FILE_OPEN:
			open();
			return TRUE;
		case IDM_FILE_SAVE:
			if (bEnabled) save();
			return TRUE;
		case IDM_FILE_SAVEAS:
			if (bEnabled) save_as();
			return TRUE;
		case IDM_FILE_RELOAD:
			if (bEnabled) reload();
			return TRUE;
		case IDM_FILE_CLOSE:
			if (bEnabled) close();
			return TRUE;
		case IDM_FILE_QUIT:
			SendMessage(hWnd, WM_CLOSE, 0, 0);
			return TRUE;
	}
	return FALSE;
}


void ProjectManager::open(){
	if (!query_saving()) return;
	if (bEnabled){					// ݂̃vWFNgĂJB
		undo_buffer.clear();		// A{͐VvWFNgǂݍ݂ɐĂׂB
		levels.clear();
		set_state(false);
	}
	OPENFILENAME ofn;
	char _filename[MAX_PATH];
	char _filetitle[64];
	ZeroMemory(&ofn, sizeof ofn);
	ZeroMemory(_filename, sizeof _filename);
	ZeroMemory(_filetitle, sizeof _filetitle);
	ofn.lStructSize			= sizeof ofn;
	ofn.hwndOwner			= hWnd;
	ofn.hInstance			= NULL;
	ofn.lpstrFilter			= "SNES ROM(*.smc)\0*.smc\0ׂẴt@C(*.*)\0*.*\0\0";
	ofn.lpstrFile			= _filename;
	ofn.nMaxFile			= sizeof _filename;
	ofn.lpstrFileTitle		= _filetitle;
	ofn.nMaxFileTitle		= sizeof _filetitle;
	ofn.Flags				= OFN_FILEMUSTEXIST;
	ofn.lpstrDefExt			= "smc";
	ofn.lpstrTitle			= "J";
	if (!GetOpenFileName(&ofn)){
		return;
	}
	strcpy_s(filename, sizeof filename, _filename);
	strcpy_s(filetitle, sizeof filetitle, _filetitle);
	load();
	return;
}

void ProjectManager::open_cmdline(LPCSTR cmdline){
	int i=0, j=0, bQuoted=FALSE, bSJIS=FALSE, c;
	while (c = cmdline[i]){
		if (bQuoted){
			if (!bSJIS && c == '"'){
				bQuoted = FALSE;
				filename[j] = '\0';
				strcpy_s(filetitle, sizeof filetitle, PathFindFileName(filename));
				if (load()) return;
			}else{
				filename[j++] = c;
				bSJIS = (!bSJIS) && ( ( c >= 0x81 && c <= 0x9F ) || ( c >= 0xE0 && c <= 0xFC ) );
			}
		}else{
			if (!bSJIS && c == '"'){
				if (j != 0){
					filename[j] = '\0';
					strcpy_s(filetitle, sizeof filetitle, PathFindFileName(filename));
					if (load()) return;
					j = 0;
				}
				bQuoted = TRUE;
			}else if (!bSJIS && c == ' '){
				if (j != 0){
					filename[j] = '\0';
					strcpy_s(filetitle, sizeof filetitle, PathFindFileName(filename));
					if (load()) return;
				}
			}else{
				filename[j++] = c;
				bSJIS = (!bSJIS) && ( ( c >= 0x81 && c <= 0x9F ) || ( c >= 0xE0 && c <= 0xFC ) );
			}
		}
		i++;
	}
	if (j != 0){
		filename[j] = '\0';
		strcpy_s(filetitle, sizeof filetitle, PathFindFileName(filename));
		if (load()) return;
	}
	return;
}

void ProjectManager::open_dropped_files(HDROP hDrop){
	if (!query_saving()) return;
	if (bEnabled){					// ݂̃vWFNgĂJB
		undo_buffer.clear();		// A{͐VvWFNgǂݍ݂ɐĂׂB
		levels.clear();
		set_state(false);
	}
	UINT nFiles = DragQueryFile(hDrop, 0xFFFFFFFF, NULL, 0);
	for (UINT i=0; i<nFiles; i++){
		DragQueryFile(hDrop, i, filename, sizeof filename);
		strcpy_s(filetitle, sizeof filetitle, PathFindFileName(filename));
		if (load()) goto open_dropped_files_finished;
	}
open_dropped_files_finished:
	DragFinish(hDrop);
	return;
}

void ProjectManager::load_file(LPCSTR _filename){
	if (!query_saving()) return;
	if (bEnabled){					// ݂̃vWFNgĂJB
		undo_buffer.clear();		// A{͐VvWFNgǂݍ݂ɐĂׂB
		levels.clear();
		set_state(false);
	}
	strcpy_s(filename, sizeof filename, _filename);
	strcpy_s(filetitle, sizeof filetitle, PathFindFileName(filename));
	load();
	return;
}

bool ProjectManager::load(){
	HANDLE hFile;
	DWORD size, size_read;
	BYTE *buf_org_without_header;
	char CartName[64];
	int sz;
	recent_files.file_opened(filename);	// ֒ǉ
	hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return false;
	size = GetFileSize(hFile, NULL);
	// wb_̗Lʂēǂݍ
	if (size == ROM_SIZE_NO_HEADER){
		bHasHeader = false;
		buf_org_without_header = new BYTE[ROM_SIZE_NO_HEADER];
		ReadFile(hFile, buf_org_without_header, ROM_SIZE_NO_HEADER, &size_read, NULL);
	}else if (size == ROM_SIZE_NO_HEADER + ROM_HEADER_SIZE){
		bHasHeader = true;
		buf_org_without_header = new BYTE[ROM_SIZE_NO_HEADER];
		ReadFile(hFile, header, ROM_HEADER_SIZE, &size_read, NULL);
		ReadFile(hFile, buf_org_without_header, ROM_SIZE_NO_HEADER, &size_read, NULL);
	}else{
		MessageBox(hWnd, "t@CTCYsłB", "G[", MB_OK | MB_ICONEXCLAMATION);
		CloseHandle(hFile);
		return false;
	}
	CloseHandle(hFile);
	// tH[}bgʂēǂݍ
	GetPrivateProfileString("ROMInfo", "CartName", "SUPER DONKEY KONG 3",
		CartName, sizeof CartName, RomInfoFileName);
	sz = strlen(CartName);
	bInterleaved = false;
	for (int i=0; i<sz; i++){
		if (buf_org_without_header[0xFFC0 + i] != CartName[i]){
			bInterleaved = true;
			break;
		}
	}
	if (bInterleaved){
		for (int i=0; i<sz; i++){
			if (buf_org_without_header[0x7FC0 + i] != CartName[i]){
				MessageBox(hWnd, "tH[}bgł܂B\r\nCartNamemFĂB",
					"G[", MB_OK | MB_ICONEXCLAMATION);
				delete[] buf_org_without_header;
				return false;
			}
		}
	}
	if (bInterleaved){
		if (buf == NULL) buf = new BYTE[ROM_SIZE_NO_HEADER];
		for (int i=0; i<0x40; i++){
			memcpy_s(buf + 0x010000*i, 0x8000, buf_org_without_header + 0x008000*i + 0x200000, 0x8000);
			memcpy_s(buf + 0x010000*i + 0x008000, 0x8000, buf_org_without_header + 0x008000*i, 0x8000);
		}
		delete[] buf_org_without_header;
	}else{
		if (buf != NULL) delete[] buf;
		buf = buf_org_without_header;
	}
	// Ĵɐ
	set_state(true);
	nChangeStep = 0;
	bChangedInrev = false;
	// Ŏqvf̏
	undo_buffer.clear();
	levels.reset();
	return true;
}

void ProjectManager::save(){
	HANDLE hFile;
	DWORD size_written;
	hFile = CreateFile(filename, GENERIC_WRITE, FILE_SHARE_READ, NULL,
		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return;
	// wb_[Ώ
	if (bHasHeader){
		WriteFile(hFile, header, ROM_HEADER_SIZE, &size_written, NULL);
	}
	// ̃tH[}bgɒď
	if (bInterleaved){
		BYTE *buf_org_without_header = new BYTE[ROM_SIZE_NO_HEADER];
		for (int i=0; i<0x40; i++){
			memcpy_s(buf_org_without_header + 0x008000*i + 0x200000, 0x8000, buf + 0x010000*i, 0x8000);
			memcpy_s(buf_org_without_header + 0x008000*i, 0x8000, buf + 0x010000*i + 0x008000, 0x8000);
		}
		WriteFile(hFile, buf_org_without_header, ROM_SIZE_NO_HEADER, &size_written, NULL);
		delete[] buf_org_without_header;
	}else{
		WriteFile(hFile, buf, ROM_SIZE_NO_HEADER, &size_written, NULL);
	}
	SetEndOfFile(hFile);
	CloseHandle(hFile);
	nChangeStep = 0;
	bChangedInrev = false;
	return;
}

void ProjectManager::save_as(){
	OPENFILENAME ofn;
	char _filename[MAX_PATH];
	char _filetitle[64];
	ZeroMemory(&ofn, sizeof ofn);
	strcpy_s(_filename, sizeof _filename, filename);
	strcpy_s(_filetitle, sizeof _filetitle, filetitle);
	ofn.lStructSize			= sizeof ofn;
	ofn.hwndOwner			= hWnd;
	ofn.hInstance			= NULL;
	ofn.lpstrFilter			= "SNES ROM(*.smc)\0*.smc\0ׂẴt@C(*.*)\0*.*\0\0";
	ofn.lpstrFile			= _filename;
	ofn.nMaxFile			= sizeof _filename;
	ofn.lpstrFileTitle		= _filetitle;
	ofn.nMaxFileTitle		= sizeof _filetitle;
	ofn.Flags				= OFN_OVERWRITEPROMPT;
	ofn.lpstrDefExt			= "smc";
	ofn.lpstrTitle			= "ۑ";
	if (!GetSaveFileName(&ofn)){
		return;
	}
	strcpy_s(filename, sizeof filename, _filename);
	strcpy_s(filetitle, sizeof filetitle, _filetitle);
	set_state(true);
	save();
	return;
}

void ProjectManager::reload(){
	if (bChangedInrev || nChangeStep!=0){
		int r = MessageBox(hWnd, "ҏWe͔j܂B낵ł?",
			"SDK3 Editor", MB_OKCANCEL | MB_ICONEXCLAMATION);
		if (r == IDCANCEL) return;
	}
	undo_buffer.clear();
	levels.clear();
	load();
	return;
}

void ProjectManager::close(){
	if (!query_saving()) return;
	undo_buffer.clear();
	levels.clear();
	set_state(false);
	return;
}

void ProjectManager::quit(){
	if (!query_saving()) return;
	destroy();
	return;
}

bool ProjectManager::query_saving(){
	if (bEnabled && (bChangedInrev || nChangeStep!=0)){
		char s[1024];
		sprintf_s(s, sizeof s, "%s ͕ύXĂ܂Bۑ܂?", filetitle);
		int r = MessageBox(hWnd, s, "SDK3 Editor", MB_ICONEXCLAMATION | MB_YESNOCANCEL);
		if (r == IDCANCEL) return false;
		if (r == IDYES){
			save();
		}
	}
	return true;
}



